################################################################################
#                     bvs_mapping.py - version 0.4                             #
################################################################################
# Copyright (C) 2013, 2017 Matthew Dyer                                        #
#   This program is free software: you can redistribute it and/or modify       #
#   it under the terms of the GNU General Public License as published by       #
#   the Free Software Foundation, either version 3 of the License, or          #
#   (at your option) any later version.                                        #
#                                                                              #
#   This program is distributed in the hope that it will be useful,            #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#   GNU General Public License for more details.                               #
#                                                                              #
#   You should have received a copy of the GNU General Public License          #
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.      #
################################################################################
# Depends on the following external python packages:                           #
#   numpy - http://www.numpy.org/                                              #
#   ase - https://wiki.fysik.dtu.dk/ase/                                       #
# They need to be installed before this script can be run                      #
################################################################################
from ase.io import read
import numpy
import math
import time

def make_grid(unit_cell,grid_spacing,center):
    """Creates a regular grid defined by a unit cell and a minimum grid spacing.
    
       Input:
           unit_cell - a 3x3 array containing Cartesian vectors for the unit 
                       cell in Angstroms
           grid_spacing - maximum spacing between two grid points along a 
                          crystallographic axis in Angstroms
    """
    #Find length of unit_cell vectors
    cell_lengths=numpy.sqrt(numpy.sum(unit_cell*unit_cell,axis=1))
    #Find number of grid points along each axis
    ngrid=numpy.array(numpy.ceil(cell_lengths/grid_spacing),numpy.int)
    print "  Number of grid points: %d %d %d" % tuple(ngrid)
    #Make grid with origin at zero
    grid=[]
    for i in range(ngrid[0]):
        x=float(i)
        for j in range(ngrid[1]):
            y=float(j)
            for k in range(ngrid[2]):
                z=float(k)
                grid.append([x,y,z])
    grid=numpy.array(grid)
    grid=grid/ngrid

    #Shift origin by half a lattice spacing in each direction
    if center=='MP':
        grid=grid+1./(2*ngrid)
    else:
        grid=grid+center
    return grid

def calculate_BVS(position,atoms,r0_iO=1.466,rho=0.37,cutoff=2.,counter_ion="O",
                  penalty=None,com=None,min_coord=1,default=1.E6):
    """Calculate the bond valance sum charge for atom at position.

       r0_iO parameters can be taken from Brese and O'Keeffe, 
       Acta Crystallographica B 47 (1991) 192; 1.466 Angstroms is for Li-O.
       The bond valence between a pair of atoms is defined as 
           s_ij = exp((r0_ij-r_ij)/rho)
       where r0_ij is the parameter given in bvs_params.
       The charge on atom i is then defined by a sum over other atoms, j
           q_i = sum_j s_ij
       where the sum takes place in real space and is limited by the value
       of cutoff x r0_ij
       Optionally, the penalty function defined in Adams, Solid State Ionics,
       177 (2006) 1625 can also be calculated be setting the variable penalty
       to the ratio of the ideal valence and the minimum coordination number.
       There is also the option to calculate the distance of the local center
       of mass of coordinated counter ions from the position. To use this 
       option set com to the cutoff in Angstroms for the coordination sphere.
    """
    tmp_atoms=atoms.copy()
    #Place position of grid point at origin
    tmp_atoms.translate(-position)
    #Work out number of repeats necessary to include all atoms within cutoff
    nrepeats=[]
    for iaxis in range(3):
        cell_length=math.sqrt(numpy.sum(atoms.cell[iaxis]**2))
        nrepeats.append(int(math.ceil(2.*r0_iO*cutoff/cell_length)))
    tmp_atoms=tmp_atoms.repeat(nrepeats)
    #Place position of grid point at center and rescale atomic positions
    scaled_positions=tmp_atoms.get_scaled_positions()
    scaled_positions=scaled_positions+[0.5,0.5,0.5]
    scaled_positions=scaled_positions%1.
    scaled_positions=scaled_positions%1.
    tmp_atoms.set_scaled_positions(scaled_positions)
    center=numpy.dot(numpy.array([0.5,0.5,0.5]),tmp_atoms.cell)
    #Calculate charge
    positions=tmp_atoms.get_positions()
    symbols=tmp_atoms.get_chemical_symbols()
    charge=0.
    if penalty:
        p_ij=0.
    if com:
        ncom=0.
        delta_com=numpy.zeros([3],numpy.float)
    for iatom in range(len(tmp_atoms)):
        if symbols[iatom]==counter_ion:
            distance=positions[iatom]-center
            distance=math.sqrt(numpy.sum(distance**2))
            if distance <= r0_iO*cutoff:
                s_ij = math.exp((r0_iO-distance)/rho)
                charge = charge + s_ij
#                charge=charge+math.exp((r0_iO-distance)/rho)
                if (penalty and s_ij > penalty):
                    p_ij = p_ij + (s_ij - penalty)**2
            if com and distance <= com:
                delta_com=delta_com+positions[iatom]-center
                ncom=ncom+1.
        
    if com:
        if ncom>=min_coord:
            delta_com=delta_com/ncom
            delta_com=math.sqrt(numpy.sum(delta_com**2))
        else:
            delta_com=default
    del tmp_atoms
    if penalty:
        return charge, penalty
	if com:
	    return charge, penalty, delta_com
    elif com:
        return charge, delta_com
    else:
        return charge

def get_grid_info(positions,tol=1.E-6):
    """Find dimensions of grid and intervals
    """
    center=positions[0]
    old_pos=center
    interval_z=positions[1]-positions[0]
    interval=interval_z
    i=1
    while (numpy.linalg.norm(interval-interval_z)<tol):
        interval=positions[i]-old_pos
        old_pos=positions[i]
        i = i+1
    n_z = i-1
    interval_y=positions[n_z]-center
    interval=interval_y
    old_pos=center
    i=n_z
    while (numpy.linalg.norm(interval-interval_y)<tol):
        interval=positions[i]-old_pos
        old_pos=positions[i]
        i = i+n_z
    n_y = i/n_z - 1
    interval_x=positions[n_y*n_z]-center
    interval=interval_x
    old_pos=center
    i=n_z*n_y
    while (numpy.linalg.norm(interval-interval_x)<tol) and i<len(positions):
        interval=positions[i]-old_pos
        old_pos=positions[i]
        i = i+n_z*n_y
    n_x = i/n_z/n_y 
    return n_x,n_y,n_z,interval_x,interval_y,interval_z

def write_grd_file(positions,charges,atoms,filename="BVS_charges.grd"):
    """Write a VESTA grd file
    """
    from ase.geometry import cell_to_cellpar
    cube_file=open(filename,"w")
    cube_file.write("BVS Map for Li-O interactions\n")
    params=cell_to_cellpar(atoms.cell)
    for param in params:
        cube_file.write("%f  " % param)
    cube_file.write("\n")
    #Find dimensions of grid and intervals
    n_x,n_y,n_z,interval_x,interval_y,interval_z=get_grid_info(positions)
    increment=numpy.dot(interval_x,atoms.cell)
    cube_file.write("%d %d %d\n" % (n_x, n_y, n_z))
    for charge in charges:
        cube_file.write("%f  " % charge)
    cube_file.write("/n")
    cube_file.close()

def write_cube_file(positions,charges,atoms,filename="BVS_charges.cub"):
    """Write a Gaussian cube file.
    """
    a0=1./0.52917721092
    cube_file=open(filename,"w")
    cube_file.write("BVS Map for Li-O interactions\n")
    cube_file.write("\n")
    center=positions[0]
    cube_file.write("%d %f %f %f\n" % (len(atoms),center[0]*a0,center[1]*a0,center[2]*a0))
    #Find dimensions of grid and intervals
    n_x,n_y,n_z,interval_x,interval_y,interval_z=get_grid_info(positions)
    cube_file.write("%d %f %f %f\n" % (n_x,interval_x[0]*a0,interval_x[1]*a0,interval_x[2]*a0))
    cube_file.write("%d %f %f %f\n" % (n_y,interval_y[0]*a0,interval_y[1]*a0,interval_y[2]*a0))
    cube_file.write("%d %f %f %f\n" % (n_z,interval_z[0]*a0,interval_z[1]*a0,interval_z[2]*a0))
    for atom in atoms:
        cube_file.write("%d %f %f %f %f\n" % (atom.number,0.,atom.position[0]*a0,atom.position[1]*a0,atom.position[2]*a0))
    for i in range(n_x):
        for j in range(n_y):
            for k in range(n_z):
                cube_file.write("%f  " % charges[i*n_y*n_z+j*n_z+k])
            cube_file.write("\n")
    cube_file.close()

def write_xplor_file(positions,charges,atoms,filename="BVS_charges.cub"):
    from ase.geometry import cell_to_cellpar
    xplor_file=open(filename,"w")
    xplor_file.write("\n")
    xplor_file.write("1\n")
    xplor_file.write("%-80s\n"%"BVS Map for Li-O interactions")
    #Find dimensions of grid and intervals
    n_x,n_y,n_z,interval_x,interval_y,interval_z=get_grid_info(positions)
    xplor_file.write("%8d%8d%8d%8d%8d%8d%8d%8d%8d\n" % (n_x,0,n_x-1,n_y,0,n_y-1,n_z,0,n_z-1))
    params=cell_to_cellpar(atoms.cell)
    for param in params:
        xplor_file.write("%12.5E" % param)
    xplor_file.write("\n")
    xplor_file.write("ZYX\n")
    for k in range(n_z):
        xplor_file.write("%8d\n" % k)
        for j in range(n_y):
            for i in range(n_x):
                index=i*n_y*n_z+j*n_z+k
                isect=j*n_x+i
                xplor_file.write("%12.5E" % charges[index])
                if isect%6==5:
                    xplor_file.write("\n")
        if (n_x*n_y)%6!=0:
            xplor_file.write("\n")
    xplor_file.write("%8d\n" % -9999)
    xplor_file.write("%12.4E %12.4E \n" % (numpy.mean(charges),numpy.std(charges)))
    xplor_file.close()

def find_maxima(positions,charges,atoms,base_name,cutoff=0.7):
    """Find and output maxima above cutoff value in map
    """
    from scipy.ndimage import label,extrema
    #Set values below cutoff to zero
    charges=charges-cutoff
    charges=numpy.clip(charges,0.,1.)
    #Find grid and redefine charges on the grid
    n_x,n_y,n_z,interval_x,interval_y,interval_z=get_grid_info(positions)
    image_array=numpy.zeros([n_x,n_y,n_z],numpy.float)
    for k in range(n_z):
        for j in range(n_y):
            for i in range(n_x):
                index=i*n_y*n_z+j*n_z+k
                image_array[i,j,k]=charges[index]
    #Find maxima
    connections=numpy.ones([3,3,3],numpy.float)
    label_array,nlabels=label(image_array,structure=connections)
    minima,maxima,min_pos,max_pos=extrema(image_array,labels=label_array,index=np.arange(1, nlabels+1))
    maxima=numpy.array(maxima)+cutoff
    #Remove maxima that are nearest neighbours or next nearest neighbours
    max_pos=numpy.array(max_pos)
    tooclose=True
    while tooclose:
        indices=[]
        for imax in range(len(maxima)):
            tmp_pos=max_pos+numpy.array([n_x/2,n_y/2,n_z/2])-max_pos[imax]
            tmp_pos=numpy.fmod(tmp_pos,[n_x,n_y,n_z])
            tmp_pos=tmp_pos-numpy.array([n_x/2,n_y/2,n_z/2])
            distances=numpy.sqrt(numpy.sum(tmp_pos**2,axis=1))
            distances[imax]=distances[imax]+10.
            if numpy.min(distances)<2.:
                indices.append(imax)
        if len(indices)>0:
            maxima=numpy.delete(maxima,indices[0])
            max_pos=numpy.delete(max_pos,indices[0],axis=0)
        else:
            tooclose=False
    #Convert positions of maxima to Cartesian coordinates
    max_positions=[]
    for position in max_pos:
        max_positions.append(positions[position[0]*n_y*n_z+position[1]*n_z+position[2]])
    #Write out values to a cif file
    outfile=open(base_name+"_maxima.txt","w")
    tmp_atoms=atoms.copy()
    for imax in range(len(max_pos)):
        outfile.write("%12.5f %12.5f %12.5f %12.5f\n" % (float(max_pos[imax][0])/float(n_x),
                      float(max_pos[imax][1])/float(n_y),float(max_pos[imax][2])/float(n_z),
                      maxima[imax]))
        tmp_atoms.extend(Atom("Li",position=max_positions[imax]))
    write(base_name+"_maxima.cif",tmp_atoms)
    outfile.close()
    

def make_map(filename,spacing,r0_iO=1.466,rho=0.37,cutoff=2.,center=[0.,0.,0.],
             base_name="BVS_charges",output="grd",penalty=None,com=None,
             min_coord=1,
             limits=None,ideal_valence=None,maximum=False,counter_ion="O"):
    """Function to produce a bond valence sum map for a given structure
    
       filename - The structure file containing the cell and atomic positions
                  for the crystal structure to be mapped. Must be a readable
                  format by ase.
       spacing - The grid spacing in Angstroms to draw the BVS map onto
       r0_iO - The bond valence sum parameter for the atom of interest with 
               oxygen (N.B. oxygen is the only anion currently available).
               Default: 1.466 Angstroms, for Li-O
       rho - The bond valence sum parameter in the denominator
       cutoff - Cutoff of atom-atom distances for bond valence sum
       center - The point on which the origin of the grid is positioned
                Default: [0.,0.,0.]
       base_name - The basis of the output filename for the map
                   Default: "BVS_charges"
       output - The output format. Currently available options are:
                "grd" - VESTA periodic grid format (default)
                "cube" - Gaussian cube format
                "xplor" - Readable by EDMA
                If these options are not present then the map will not be 
                written to file
       limits - The upper and lower limits of the map data to print out.
                Default: None - no limits are set
                The limits object should be a list object containing floats
                for the lower and then upper limits
                   e.g. [0.7,1.3] 
       ideal_valence - If the ideal valence is specified, then the absolute
                       difference from this value (the BV mismatch 
                       Adams, J. Power Sources, 159 (2006) 200) is calculated
                       and written out along with the BVS map
       maximum - Instead of outputting the BV mismatch, the function 
                 ideal_valence - mismatch is output, which should have maxima 
                 when the BVS is close to ideal.
                 If maximum is defined as a float, then the positions and values
                 of all maxima above that value will be found and output to file
       counter_ion - Counter ion for BVS calculations. Default is oxygen
       penalty - If set then the penalty function defined in Adams, Solid State
                 Ionics, 177 (2006) 1625 is calculated. Set the variable penalty
                 to the ratio of the ideal valence and the minimum coordination 
                 number.
       com - There is also the option to calculate the distance of the local 
             center of mass of coordinated counter ions from the position. To 
             use this option set com to the cutoff in Angstroms for the 
             coordination sphere.
       min_coord - If the coordination sphere used for calculation of the local
                   center of mass contains fewer coordinating atoms than this 
                   value, then com is returned as a large number (default 1E6)
    """
    start_time=time.time()
    #Read in the atoms object
    atoms=read(filename)
    #Create the grid
    grid=make_grid(atoms.cell,spacing,center)
    grid=numpy.dot(grid,atoms.cell)
    #Calculate bond valence sum charges at each point
    charges=[]
    if penalty:
        penalty_function=[]
    if com:
        com_function=[]
    for position in grid:
        if penalty and com:
            charge,p_ij,delta_com=calculate_BVS(position,atoms,r0_iO,rho,cutoff,
                                     counter_ion,penalty,com,min_coord=min_coord)
            charges.append(charge)
            penalty_function.append(p_ij)
            com_function.append(delta_com)
        elif penalty:
            charge,p_ij=calculate_BVS(position,atoms,r0_iO,rho,cutoff,
                                     counter_ion,penalty)
            charges.append(charge)
            penalty_function.append(p_ij)
        elif com:
            charge,delta_com=calculate_BVS(position,atoms,r0_iO,rho,cutoff,
                                     counter_ion,com=com,min_coord=min_coord)
            charges.append(charge)
            com_function.append(delta_com)
        else:
            charges.append(calculate_BVS(position,atoms,r0_iO,rho,cutoff,
                                     counter_ion))
    
    #If limits are set then apply them
    if limits:
        for icharge in range(len(charges)):
            if charges[icharge]<limits[0] or charges[icharge]>limits[1]:
                charges[icharge]=0.

    #Output file written if needed
    if output.lower().find("cub")!=-1:
        write_cube_file(grid,charges,atoms,filename=base_name+".cub")
    if output.lower().find("grd")!=-1:
        write_grd_file(grid,charges,atoms,filename=base_name+".grd")
    if output.lower().find("xplor")!=-1:
        write_xplor_file(grid,charges,atoms,filename=base_name+".xplor")

    #If ideal valence is set then calculate the absolute difference away from it
    if ideal_valence:
        mismatch=numpy.array(charges)
        mismatch=numpy.abs(mismatch-ideal_valence)
        suffix="_mismatch"
        #Apply penalty function if wanted
        if penalty:
            penalty_function = numpy.array(penalty_function)
            mismatch=mismatch+penalty_function
        if com:
            suffix="_sumcom"
            com_function = numpy.array(com_function)
            mismatch=mismatch+com_function
        #Switch to 1 - mismatch to give maxima at ideal valence
        if maximum:
            mismatch=ideal_valence-mismatch
            mismatch=numpy.clip(mismatch,0.,ideal_valence)
            suffix="_maxima"
            if type(maximum)==float:
                find_maxima(grid,mismatch,atoms,base_name,cutoff=maximum)
        if output.lower().find("cub")!=-1:
            write_cube_file(grid,mismatch,atoms,filename=base_name+suffix+".cub")
        if output.lower().find("grd")!=-1:
            write_grd_file(grid,mismatch,atoms,filename=base_name+suffix+".grd")
        if output.lower().find("xplor")!=-1:
            write_xplor_file(grid,mismatch,atoms,filename=base_name+suffix+".xplor")

    #If displacement of centre of mass is calculated, print it out
    if com:
        suffix="_com"
        if output.lower().find("cub")!=-1:
            write_cube_file(grid,com_function,atoms,filename=base_name+suffix+".cub")
        if output.lower().find("grd")!=-1:
            write_grd_file(grid,com_function,atoms,filename=base_name+suffix+".grd")
        if output.lower().find("xplor")!=-1:
            write_xplor_file(grid,com_function,atoms,filename=base_name+suffix+".xplor")

    #Print out total time taken
    print "Total time taken = %.4f s" % (time.time()-start_time)
